import numpy as np 
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed
import mne
import scipy
from scipy.signal import *
from scipy.stats import gamma
import scipy.special as sc

def reorder(ori, shuffled):
    return np.where(shuffled.reshape(-1, 1) == ori.reshape(1, -1))[1]


def parallel_process(array, function, n_jobs=16, use_kwargs=False, front_num=3):
    """
        A parallel version of the map function with a progress bar. 

        Args:
            array (array-like): An array to iterate over.
            function (function): A python function to apply to the elements of array
            n_jobs (int, default=16): The number of cores to use
            use_kwargs (boolean, default=False): Whether to consider the elements of array as dictionaries of 
                keyword arguments to function 
            front_num (int, default=3): The number of iterations to run serially before kicking off the parallel job. 
                Useful for catching bugs
        Returns:
            [function(array[0]), function(array[1]), ...]
    """
    #We run the first few iterations serially to catch bugs
    if front_num > 0:
        front = [function(**a) if use_kwargs else function(a) for a in array[:front_num]]
    #If we set n_jobs to 1, just run a list comprehension. This is useful for benchmarking and debugging.
    if n_jobs==1:
        return front + [function(**a) if use_kwargs else function(a) for a in tqdm(array[front_num:])]
    #Assemble the workers
    with ProcessPoolExecutor(max_workers=n_jobs) as pool:
        #Pass the elements of array into function
        if use_kwargs:
            futures = [pool.submit(function, **a) for a in array[front_num:]]
        else:
            futures = [pool.submit(function, a) for a in array[front_num:]]
        kwargs = {
            'total': len(futures),
            'unit': 'it',
            'unit_scale': True,
            'leave': True
        }
        #Print out the progress as tasks complete
        for f in tqdm(as_completed(futures), **kwargs):
            pass
    out = []
    #Get the results from the futures. 
    for i, future in enumerate(futures):
        try:
            out.append(future.result())
        except Exception as e:
            out.append(e)
    return front + out


def read_raw(raw_path):
    """Read raw data from an EDF file
    
    Parameters
    ----------
    raw_path : str
        Path to the EDF file
    
    Returns
    -------
    data : numpy array
    channels : numpy array"""
    raw_path = validate_type(raw_path, 'raw_path', str)
    raw = mne.io.read_raw_edf(raw_path)
    raw_channels = raw.info['ch_names']
    data = []

    for raw_ch in raw_channels:
        ch_data = raw.get_data(raw_ch) * 1E6
        data.append(ch_data)

    data = np.squeeze(data)
    channels = np.array(raw_channels)
    return data, channels


def preprocess(raw, sample_freq, filter_freq, rp=0.5, rs=93, space=0.5):
    """Preprocess the raw data
    Parameters
    ----------
    raw : numpy array like of float, shape (n_samples,)
        raw data
    sample_freq : float
        sample frequency
    filter_freq : list of float, shape (2,)
        filter frequency
    rp : float | 0.5
        max loss in passband (dB)
    rs : float | 93
        min attenuation in stopband (dB)
    space : float

    Returns
    -------
    filtered : array
        filtered data
    """
    # check raw data
    raw = validate_type(raw, 'raw', np.ndarray)
    if raw.ndim != 1:
        raise ValueError('raw must be a 1D array')
    # check sample frequency
    sample_freq = validate_param(sample_freq, 'sample_freq', (int, float))
    # check filter frequency
    filter_freq = validate_filter_freq(filter_freq)

    nyq = sample_freq / 2
    # MGNM
    if filter_freq[1] >= .99 * nyq:
        filter_freq[1] = nyq * .99
    low, high = filter_freq

    scale = 0
    while 0 < low < 1:
        low *= 10
        scale += 1
    low = filter_freq[0] - (space * 10 ** (-1 * scale))

    scale = 0
    while high < 1:
        high *= 10
        scale += 1
    high = filter_freq[1] + (space * 10 ** (-1 * scale))  
    stop_freq = [low, high]
    ps_order, wst = cheb2ord([filter_freq[0] / nyq, filter_freq[1] / nyq], [stop_freq[0] / nyq, stop_freq[1] / nyq], rp,
                                rs)
    z, p, k = cheby2(ps_order, rs, wst, btype='bandpass', analog=0, output='zpk')
    sos = zpk2sos(z, p, k)    

    filtered = sosfilt(sos, raw)
    filtered = sosfilt(sos, np.flipud(filtered))
    filtered = np.flipud(filtered)

    return filtered


def validate_type(var, var_name, var_type):
    """Validate the type of a variable"""
    if not isinstance(var, var_type):
        raise TypeError('{} must be a {}'.format(var_name, var_type))
    return var


def validate_param(var, var_name, var_type):
    """Validate the type and value of a variable"""
    var = validate_type(var, var_name, var_type)
    if var < 0:
        raise ValueError('{} must not be a negative {}'.format(var_name, var_type))
    return var


def validate_filter_freq(filter_freq):
    """Validate filter frequency
    Parameters
    ----------
    filter_freq : array
        filter frequency
    Returns
    -------
    filter_freq : array
        filter frequency
    """
    filter_freq = validate_type(filter_freq, 'filter_freq', (list, tuple, np.ndarray))
    if len(filter_freq) != 2:
        raise ValueError('filter_freq must have two elements')
    if not all(isinstance(f, (int, float)) for f in filter_freq):
        raise TypeError('filter_freq must be a list of int or float')   
    if filter_freq[0] >= filter_freq[1]:
        raise ValueError('filter_freq[0] must be smaller than filter_freq[1]')
    if filter_freq[0] <= 0:
        raise ValueError('filter_freq[0] must be positive')
    return filter_freq


def compute_rms(filtered, sample_freq, window_size=6*1e-3, detector=None):
    """Compute RMS
    Parameters
    ----------
    filtered : numpy array like of float, shape (n_samples,)
        filtered data
    sample_freq : float | int
        sample frequency
    window_size : float | int
        window size
    detector : str
        detector type
    Returns
    -------
    rms : array of float, shape (n_samples,)
        RMS
    """
    # check filtered data
    filtered = validate_type(filtered, 'filtered', np.ndarray)
    if filtered.ndim != 1:
        raise ValueError('filtered must be a 1D array')
    # check sample frequency
    sample_freq = validate_param(sample_freq, 'sample_freq', (int, float))
    # check window size 
    window_size = validate_param(window_size, 'window_size', (int, float))
    # check detector
    if detector is not None:
        detector = validate_type(detector, 'detector', str)
        if detector not in ['RMS', 'MNI']:
            raise ValueError('detector must be RMS or MNI')

    # compute RMS
    if detector == 'RMS':
        rms_window = round(window_size * sample_freq)
    elif detector == 'MNI':
        rms_window = round(0.002 * sample_freq)
    
    if rms_window % 2 == 0:
        rms_window += 1
    
    temp = np.square(filtered)
    temp = lfilter(np.ones(rms_window), 1, temp, axis=0) / rms_window
    rms = np.zeros(len(temp))
    rms[:int(-np.floor(rms_window / 2))] = temp[int(np.floor(rms_window / 2)):]
    rms = np.sqrt(rms)

    return rms


def compute_wavelet_entropy(sample_freq, epoch_samples, low, high, seed, dev_cycles=3):
    """Compute wavelet entropy
    Parameters
    ----------
    sample_freq : int | float
        sample frequency
    epoch_samples : int | float
        epoch samples
    low : int | float
        low frequency
    high : int | float
        high frequency
    dev_cycles : int | float | 3 | optional
        number of cycles
    seed: int | None
        seed for random number generator
    Returns
    -------
    wavelet_entropy : int | float
        wavelet entropy"""
    #check inputs
    sample_freq = validate_param(sample_freq, 'sample_freq', (int, float))
    epoch_samples = validate_param(epoch_samples, 'epoch_samples', (int, float))
    low = validate_param(low, 'low', (int, float))
    high = validate_param(high, 'high', (int, float))
    dev_cycles = validate_param(dev_cycles, 'dev_cycles', (int, float))
    
    np.random.seed(seed)
    wavelet_entropy_max = np.zeros((100,1))
    for i in range(0, 100):
        segment = np.random.rand(epoch_samples, 1)
        auto_corr = correlate(segment, segment, mode='full') / np.sum(segment ** 2)
        w_coef = gabor_transform_wait(auto_corr, sample_freq, low, high, dev_cycles)
        prob_energy = np.mean(w_coef ** 2, 1) / np.sum(np.mean(w_coef ** 2, 1))
        wavelet_entropy_max[i] = -np.sum(prob_energy * np.log(prob_energy))

    del segment, prob_energy, w_coef

    return np.median(wavelet_entropy_max)


def gabor_transform_wait(pv_signal, sample_rate, min_freq, max_freq, dev_cycles=3, magnitudes=1,
                         squared_mag=0,
                         band_ave=0, phases=0, time_step=[]):
    """This function calculates the Wavelet Transform using a Gaussian modulated window (Gabor Wavelet).
    Parameters
    ----------
    pv_signal : array
        Signal to process
    sample_rate : float | int
        sample rate
    min_freq : float | int
        Min frequency (in Hz) to process from
    max_freq : float | int
        Max frequency (in Hz) to process to
    dev_cycles : float | int
         the number of cycles to include for the wavelet transform at every frequency.
    magnitudes : int|float | 1 | optional
        Set to 1 (default) if the magnitudes of the coefficients must be returned; 0 for analytic values (complex values).
    squared_mag : int|float | 0 | optional
        Set to 1 if the magnitudes of the coefficients divided by the squared of the corresponding scale must by power to 2
    band_ave : int|float | 0 | optional
        Set to 1 if instead of returning a matrix with all values
        in the time-frequency map, the function returns just a vector with the
        average along all the frequency scales for each time moment.
    phases : int | 0 | optional
        Set to 1 if the phases of the coefficients must be returned; 0 for analytic values (complex values).
    time_step : array|list | [] | optional
        time step between values that are going to be kept in the output matrix
        Each time moment is the average of the previous values according to the size of the window defined by this parameter.
    
    Returns
    -------
    w_coef : array
        wavelet transform output matrix (n_freqs x n_samples) Time in rows, Frequency in columns
    """
    # check inputs
    pv_signal = validate_type(pv_signal, 'pv_signal', np.ndarray)
    sample_rate = validate_param(sample_rate, 'sample_rate', (int, float))
    min_freq = validate_param(min_freq, 'min_freq', (int, float))
    max_freq = validate_param(max_freq, 'max_freq', (int, float))
    dev_cycles = validate_param(dev_cycles, 'dev_cycles', (int, float))
    magnitudes = validate_param(magnitudes, 'magnitudes', (int, float))
    squared_mag = validate_param(squared_mag, 'squared_mag', (int, float))
    band_ave = validate_param(band_ave, 'band_ave', (int, float))
    phases = validate_param(phases, 'phases', (int, float))
    time_step = validate_type(time_step, 'time_step', (list, tuple, np.ndarray))

    pv_signal = pv_signal[:]
    freq_seg = len(list(range(min_freq, max_freq, 5)))
    freq_step = int((max_freq - min_freq) / (freq_seg - 1))
    freq_axis = np.arange(min_freq, max_freq + freq_step, freq_step)
    freq_axis = np.flip(freq_axis)
    if len(pv_signal) % 2 == 0:
        pv_signal = pv_signal[0:len(pv_signal) - 1]
    
    time_axis = (np.arange(0, len(pv_signal))) / sample_rate
    s_len = len(time_axis)
    s_half_len = (s_len // 2) + 1

    w_axis = (2 * np.pi / s_len) * (np.arange(0, s_len))
    w_axis = w_axis * sample_rate
    w_axis_half = w_axis[:s_half_len]

    if not time_step:
        sample_ave = 1
    else:
        sample_ave = [round(ts * sample_rate) for ts in time_step]
        if all(x < 1 for x in sample_ave):
            sample_ave = 1

    sample_ave_filt = []
    if sample_ave > 1:
        ind_samp = list(range(1, len(time_axis), sample_ave))
        time_axis = [time_axis[i] for i in ind_samp]
        sample_ave_filt = np.ones(sample_ave, 1)

    signal_fft = scipy.fft.fft(pv_signal, len(pv_signal), axis=0)
    signal_fft = signal_fft.reshape((-1, 1))

    gabor_wt = np.zeros((len(freq_axis), len(time_axis)), dtype=complex)
    freq_ind = 0
    for freq_counter in freq_axis:
        dev_sec = (1 / freq_counter) * dev_cycles 
        win_fft = np.zeros((s_len, 1))
        tmp1 = (w_axis_half - 2 * np.pi * freq_counter) ** 2
        tmp2 = dev_sec ** 2
        tmp = np.exp(-0.5 * tmp1 * tmp2)
        win_fft[:s_half_len] = np.reshape(tmp, (len(tmp), 1))
        win_fft = win_fft * np.sqrt(s_len) / np.linalg.norm(win_fft)
        freq_ind += 1

        if sample_ave > 1:
            gabor_tmp = np.zeros((len(signal_fft) + (sample_ave - 1), 1), dtype=complex)
            gabor_tmp[sample_ave:] = scipy.fft.ifft(signal_fft * win_fft, axis=0) / np.sqrt(dev_sec)

            if magnitudes:
                gabor_tmp = abs(gabor_tmp)
            if squared_mag:
                gabor_tmp **= 2

            gabor_tmp[:(sample_ave - 1)] = np.flipud(gabor_tmp[(sample_ave + 1):(2 * sample_ave - 1)])
            gabor_tmp = filter(sample_ave_filt, 1, gabor_tmp) / sample_ave
            gabor_tmp = gabor_tmp[sample_ave:]

            gabor_wt[freq_ind - 1] = gabor_tmp[ind_samp]

        else:
            tmp_ifft = scipy.fft.ifft(signal_fft * win_fft, axis=0) / np.sqrt(dev_sec)
            gabor_wt[freq_ind - 1] = tmp_ifft.reshape((1, len(tmp_ifft)))

    if sample_ave > 1:
        return gabor_wt

    if phases != 0:
        gabor_wt = np.angle(gabor_wt)
        return gabor_wt

    if magnitudes != 1:
        return gabor_wt

    gabor_wt = abs(gabor_wt)

    if squared_mag != 0:
        gabor_wt **= 2

    if band_ave != 0:
        gabor_wt = np.mean(gabor_wt, 2)
        gabor_wt = np.flipud(gabor_wt)
        time_axis = []
        freq_axis = np.fliplr(freq_axis)

    return gabor_wt


def pos_baseline(baseline_window, rms, epoch_time, sample_freq, thr_perc):
    """Calculate baseline for positive peaks.

    Parameters
    ----------
    baseline_window : array
        baseline window
    rms : array
        rms
    epoch_time : int|float
        epoch time
    sample_freq : int|float
        sample frequency
    thr_perc : int|float
        threshold percentage
    Returns
    ------- 
    res : array
        baseline
    """
    # check inputs
    baseline_window = validate_type(baseline_window, 'baseline_window', np.ndarray)
    if baseline_window.ndim != 2 and baseline_window.shape[1] != 1:
        raise ValueError('baseline_window must be a 2D array with shape (n_samples, 1)')
    if not np.equal(baseline_window, 0).any() or not np.equal(baseline_window, 1).any():
        raise ValueError('baseline_window contains only 0 and 1')
    rms = validate_type(rms, 'rms', np.ndarray)
    epoch_time = validate_param(epoch_time, 'epoch_time', (int, float))
    sample_freq = validate_param(sample_freq, 'sample_freq', (int, float))
    thr_perc = validate_param(thr_perc, 'thr_perc', (int, float))

    # calculate baseline
    window_thrd = round(epoch_time * sample_freq)
    bw_tmp = np.zeros((len(baseline_window)+2, 1))
    bw_tmp[1:-1] = baseline_window
    win_baseline = np.diff(bw_tmp, axis=0)
    win_baseline_up = np.where(win_baseline == 1)[0] + 1
    win_baseline_down = np.where(win_baseline == -1)[0]
    
    idx_ini = []
    for i in range(len(win_baseline_up)):
        idx_add = list(range(win_baseline_up[i], win_baseline_down[i], window_thrd))
        idx_ini.extend(idx_add)
    idx_ini = np.array(idx_ini)
    idx_end = idx_ini + window_thrd - 1
    idx_rem = idx_end <= len(rms)
    idx_ini = idx_ini[idx_rem]
    idx_end = idx_end[idx_rem]       

    res = np.zeros(rms.shape)
    for i in range(len(idx_ini)):
        sect = np.sort(rms[int(idx_ini[i]) - 1:int(idx_end[i])])
        gam_params = fb_gamfit(sect)
        percent = fb_gamcdf(sect, gam_params[0], gam_params[1])
        p_idxs = np.where(percent <= thr_perc)
        pi = p_idxs[0][-1]
        res[int(idx_ini[i]) - 1:int(idx_end[i])] = sect[pi]

    return res


def neg_baseline(rms, min_win, sample_freq, epo_CHF, per_CHF):
    """Calculate baseline for negative peaks.

    Parameters
    ----------
    rms : array
    min_win : int
        minimum HFO Time window
    sample_freq : int|float 
        sample frequency
    epo_CHF : int|float
        Continuous High Frequency Epoch
    per_CHF : int|float
        Continuous High Frequency Percentile Threshold
    
    Returns
    -------
    res : array
        baseline
    """
    # check inputs
    rms = validate_type(rms, 'rms', np.ndarray)
    min_win = validate_param(min_win, 'min_win', (int, float))
    sample_freq = validate_param(sample_freq, 'sample_freq', (int, float))
    epo_CHF = validate_param(epo_CHF, 'epo_CHF', (int, float))
    per_CHF = validate_param(per_CHF, 'per_CHF', (int, float))

    # calculate baseline
    min_win = min_win * sample_freq
    window_chf = round(epo_CHF * sample_freq)
    idx_ini = np.arange(0, len(rms), window_chf)
    idx_end = idx_ini + window_chf - 1
    idx_rem = idx_end <= len(rms)
    idx_ini = idx_ini[idx_rem]
    idx_end = idx_end[idx_rem]

    res = np.zeros(rms.shape)
    for i in range(len(idx_ini)):
        sect = np.sort(rms[int(idx_ini[i]):int(idx_end[i]) + 1], axis=0)
        thres_last = np.max(sect)
        while True:
            if sum(abs(sect)) == 0:
                break
            # MLE of params of gamma distribution fit to the data(sect), return gam_params = (shape, scale)
            if np.all(sect != 0):
                gam_params = gamma.fit(sect, floc=0)
            else:
                gam_params = gamma_fit(sect)
            percent = gamma.cdf(sect, gam_params[0], gam_params[1], gam_params[2])
            p_idxs = np.where(percent <= per_CHF)
            thres_new = sect[p_idxs[0][-1]]
            energy_over = np.zeros(len(sect)+2)
            energy_over[1:-1] = sect >= thres_new
            jumps = np.diff(energy_over, axis=0)
            jump_up = np.where(jumps == 1)[0]
            jump_down = np.where(jumps == -1)[0] - 1
            dist = jump_down - jump_up

            select = np.where(dist > min_win)[0]
            if select.size == 0:
                break
            ini = jump_up[select]
            end = jump_down[select]

            for ii in range(len(ini)):
                sect[ini[ii]: end[ii] + 1] = 0
                sect = np.sort(sect,axis=0)

            thres_last = thres_new

        res[int(idx_ini[i]):int(idx_end[i]) + 1] = thres_last

    return res


def gampdf(x, a, b):
    """Gamma probability density function.

    Parameters
    ----------
    x : array
        input values
    a : int|float
        parameter of the gamma distribution
    b : int|float
        parameter of the gamma distribution

    Returns
    -------
    res : array
        the probability density function evaluated at x
    """
    # check inputs
    x = validate_type(x, 'x', np.ndarray)
    a = validate_type(a, 'a', (int, float))
    b = validate_type(b, 'b', (int, float))
    # calculate
    sz = len(x)
    pdf = np.zeros(sz)

    k = np.where(np.logical_or(a <= 0, b <= 0, np.isnan(x)))[0]
    if k.size != 0:
        pdf[k] = np.nan

    k = np.where(np.logical_and(x > 0, 0 < a <= 1 and b > 0))[0]
    if k.size != 0:
        if np.isscalar(a) and np.isscalar(b):
            pdf[k] = (x[k] ** (a-1)) * np.exp(-x[k] / b) / sc.gamma(a) / (b ** a)
        else:
            pdf[k] = (x[k] ** (a[k]-1)) * np.exp(-x[k] / b[k]) / sc.gamma(a[k]) / (b[k] ** a[k])

    k = np.where(np.logical_and(x > 0, a > 1 and b > 0))[0]
    if k.size != 0:
        if np.isscalar(a) and np.isscalar(b):
            pdf[k] = np.exp(-a * np.log(b) + (a-1) * np.log(x[k]) - x[k] / b - sc.gammaln(a))
        else:
            pdf[k] = np.exp(-a[k] * np.log(b[k]) + (a[k]-1) * np.log(x[k]) - x[k] / b[k] - sc.gammaln(a[k]))

    return pdf

            
def gamlike(p, x):
    """Negative log-likelihood function for the Gamma distribution over vector x, with the given parameters p[0] and p[1].

    Parameters
    ----------
    p : array
        parameters of the gamma distribution (shape, scale)
    x : array
        input values

    Returns
    -------
    res : array
        the negative log-likelihood function evaluated at x with the given parameters
    """
    # check inputs
    p = validate_type(p, 'p', (list, np.ndarray))
    if len(p) != 2:
        raise ValueError('p must be a list or array with 2 elements')
    p[0] = validate_type(p[0], 'p[0]', (int, float))
    p[1] = validate_type(p[1], 'p[1]', (int, float))
    x = validate_type(x, 'x', np.ndarray)

    with np.errstate(divide='ignore'): 
        return -np.sum(np.log(gampdf(x, p[0], p[1])))


def gamfit_search(a, avg, x):
    """Search for the best fit of gamma distribution.

    Parameters
    ----------
    a : int|float
    avg : int|float
    x : array

    Returns
    -------
    res : array
    """
    # check inputs
    a = validate_type(a, 'a', (int, float))
    avg = validate_type(avg, 'avg', (int, float))
    x = validate_type(x, 'x', np.ndarray)  
    # calculate
    return -gamlike([a, avg / a], x)


def nmsmax(func, x, stopit=[], savit=[], *args):
    """Nelder-Mead simplex method for direct search optimization.

    Parameters
    ----------
    func : function
        function to be optimized
    x : list|np.ndarray 
        initial guess
    stopit : list|np.ndarray 
        stopping criteria
    saveit : list|np.ndarray 
        saving criteria
    args : array|tuple|list
        additional arguments for the function

    Returns
    -------
    x[0] : array
        vector yielding largest function value found
    fmax : float   
        function value at x
    nf : int
        number of function evaluations
    """
    x = validate_type(x, 'x', (list, np.ndarray))
    stopit = validate_type(stopit, 'stopit', (list, np.ndarray))
    savit = validate_type(savit, 'savit', (list, np.ndarray))

    x0 = x[:]
    n = len(x0)

    stopit = [1e-3, np.inf, np.inf, 0, 0]
    tol, trace = stopit[0], stopit[4]

    V = np.hstack((np.zeros((n, 1)), np.eye(n)))
    f = np.zeros((n + 1, 1))
    V[:, 0] = x0
    f[0] = func(1, *args)
    fmax_old = f[0]

    k, m = 0, 0

    scale = np.max((np.linalg.norm(np.array(x), np.inf), 1))
    alpha = scale / (n * np.sqrt(2)) * np.array([np.sqrt(n + 1) - 1 + n, np.sqrt(n + 1) - 1])
    V[:, 1:n + 1] = x0 + alpha[1] * np.ones((n, 1)) * np.ones((1, n))
    for i in range(1, n + 1):
        V[i - 1, i] = x0[i - 1] + alpha[0]
        x[:] = V[:, i]
        f[i] = func(x[0], *args)
    nf = n + 1
    temp = np.sort(f)
    idx = np.where(temp)[0]
    idx = idx[::-1]
    f = f[idx]
    V = V[:, idx]

    alpha, beta, gamma = 1, 1 / 2, 2

    while True:
        k += 1
        fmax = f[0]
        if fmax > fmax_old:
            if len(savit) != 0:
                x[:] = V[:, 0]
        fmax_old = fmax
        # three stopping tests
        # Stopping Test 1 - f reached target value
        if fmax >= stopit[2]:
            break
        # Stopping Test 2 - too many f-evals
        if nf >= stopit[1]:
            break
        # Stopping Test 3 - converged
        v1 = V[:, 0]
        size_simplex = np.linalg.norm(V[:, 1:n + 1] - np.tile(v1, n), 1) / np.max((1, np.linalg.norm(v1, 1)))
        if size_simplex <= tol:
            break

        # One step of the Nelder-Mead simplex algorithm
        vbar = np.transpose(np.sum(np.transpose(V[:, 0:n]) / n))
        vr = (1 + alpha) * vbar - alpha * V[:, n]
        x[:] = vr
        fr = func(x[0], *args)
        nf += 1
        vk, fk = vr, fr
        if fr > f[n - 1]:
            if fr > f[0]:
                ve = gamma * vr + (1 - gamma) * vbar
                x[:] = ve
                fe = func(x[0], *args)
                nf += 1
                if fe > f[0]:
                    vk, fk = ve, fe
        else:
            vt = V[:, n]
            ft = f[n]
            if fr > ft:
                vt, ft = vr, fr
            vc = beta * vt + (1 - beta) * vbar
            x[:] = vc
            fc = func(x[0], *args)
            nf += 1
            if fc > f[n - 1]:
                vk, fk = vc, fc
            else:
                for j in range(1, n):
                    V[:, j] = (V[:, 0] + V[:, j]) / 2
                    x[:] = V[:, j]
                    f[j] = func(x[0], *args)
                nf += (n - 1)
                vk = (V[:, 0] + V[:, n]) / 2
                x[:] = vk
                fk = func(x[0], *args)
                nf += 1
        V[:, n] = vk
        f[n] = fk
        temp = np.sort(f)
        idx = np.where(temp)[0]
        idx = idx[::-1]
        f = f[idx]
        V = V[:, idx]

    x[:] = V[:, 0]
    return x[0], fmax[0], nf


def fb_gamfit(x):
    """the maximumlikelihood estimator for the Gamma distribution for x

    Parameters
    ----------
    x : array
        data

    Returns
    ------- 
    res : list of float, length 2
        [a, b] where a is the shape and b is the scale
    """
    # check inputs
    x = validate_type(x, 'x', np.ndarray)
    # fit gamma distribution
    avg = np.mean(x)
    a, _, _ = nmsmax(gamfit_search, [1], [], [], avg, x)
    b = a / avg
    res = [a, 1 / b]
    return res


def fb_gamcdf(x, a, b):
    """the cumulative distribution function for the Gamma distribution for x given a and b

    Parameters
    ----------
    x : array
        data
    a : float | int
        shape
    b : float | int
        scale

    Returns
    -------
    cdf : array
        cumulative distribution function values for x 
    """

    # check inputs
    x = validate_type(x, 'x', np.ndarray)
    a = validate_type(a, 'a', (int, float))
    b = validate_type(b, 'b', (int, float))

    # calculate cdf
    sz = len(x)
    cdf = np.zeros(sz)

    k = np.where(np.logical_or(a <= 0, b <= 0, np.isnan(x)))[0]
    if k.size != 0:
        cdf[k] = np.nan

    k = np.where(np.logical_and(x > 0, a > 0 and b > 0))[0]
    if k.size != 0:
        if np.isscalar(a) and np.isscalar(b):
            cdf[k] = sc.gammainc(a, x[k]/b)
        else:
            cdf[k] = sc.gammainc(a[k], x[k]/b[k])
    return cdf


def gamma_fit(x):
    n = len(x)
    freq = 1
    scalex = np.sum(freq*x) / n
    x = x / scalex
    xbar = 1
    s2 = np.sum(freq*(x-xbar)**2) / n
    s2 = s2 * n / (n - 1)
    ahat = xbar ** 2 / s2
    bhat = s2 / xbar
    parmhat = [ahat, 0, bhat*scalex]
    # parmhat[2] = parmhat[2] * scalex
    return parmhat





    



import numpy as np
from scipy.signal import *
from .utils import *

class STEDetector():
    
    def __init__(self, sample_freq, filter_freq=[80, 500], 
                rms_window=3 * 1e-3, min_window=6 * 1e-3, min_gap=10 * 1e-3, 
                epoch_len=600, min_osc=6, rms_thres=5, peak_thres=3,
                n_jobs=32, use_kwargs=True, front_num=1):
        """Initialize the RMS detector
        sample_freq : float | int
            Sampling frequency of the data in Hz
        filter_freq : list of float or int, length 2, default [80, 500] Hz
            Filter freqs in Hz
        rms_window : float | int, default 3 * 1e-3 s
            RMS window time in seconds(s)
        min_window : float | int, default 6 * 1e-3 s
            Minimum window time for an HFO in seconds(s)
        min_gap : float | int, default 10 * 1e-3 s
            Minimum distance time between two HFO candidates in seconds(s)
        epoch_len : float | int, default 600 s
            Cycle time in seconds(s)
        min_osc : float | int, default 6
            Minimum oscillations per interval
        rms_thres : float | int, default 5
            RMS threshold in standard deviation
        peak_thres : float | int, default 3
            threshold for peak detection
        n_jobs : int, default 32
            Number of jobs to run in parallel
        use_kwargs : bool, default True
            Whether to pass additional arguments to the map function
        front_num : int, default 1
            Number of jobs to run in parallel at the front end
        """
        sample_freq = validate_param(sample_freq, 'sample_freq', (int, float))
        filter_freq = validate_filter_freq(filter_freq)
        if filter_freq[1] > sample_freq/2:
            raise ValueError('filter_freq[1] must be less than sample_freq/2')
        rms_window = validate_param(rms_window, 'rms_window', (int, float))
        min_window = validate_param(min_window, 'min_window', (int, float))
        min_gap = validate_param(min_gap, 'min_gap', (int, float))
        epoch_len = validate_param(epoch_len, 'epoch_len', (int, float))  
        min_osc = validate_param(min_osc, 'min_osc', (int, float))
        rms_thres = validate_param(rms_thres, 'rms_thres', (int, float))
        peak_thres = validate_param(peak_thres, 'peak_thres', (int, float))
        n_jobs = validate_param(n_jobs, 'n_jobs', int)
        use_kwargs = validate_param(use_kwargs, 'use_kwargs', bool)
        front_num = validate_param(front_num, 'front_num', int)
        
        self.sample_freq = sample_freq
        self.filter_freq = filter_freq
        self.rms_window = rms_window
        self.min_window = min_window
        self.min_gap = min_gap
        self.epoch_len = epoch_len
        self.min_osc = min_osc
        self.rms_thres = rms_thres
        self.peak_thres = peak_thres
        self.n_jobs = n_jobs
        self.use_kwargs = use_kwargs
        self.front_num = front_num


    def detect_edf(self, edf_path):
        """detect HFOs from an EDF file
        
        Parameters
        ----------
        edf_path : str
            Path to the EDF file
            
        Returns
        -------
        channel_names : numpy array of str, shape (n_channels,)
        HFOs : numpy array of int, shape (n_channels, n_HFOs, 2)"""
        
        edf_path = validate_type(edf_path, 'edf_path', str)
        raw, channels = read_raw(edf_path)
        return self.detect_multi_channels(data=raw, channels=channels)


    def detect_multi_channels(self, data, channels, filtered=False):
        """detect HFOs from numpy arrays of multi channels parrallelly
        
        Parameters
        ----------
        data : numpy array like of float, shape (n_channels, n_samples)
        channels : numpy array like of str, shape (n_channels,)
        filtered : bool, default False, 
            Whether the data is already filtered
        
        Returns
        -------
        channel_names : numpy array of str, shape (n_channels,)
        HFOs : numpy array of int, shape (n_channels, n_HFOs, 2)"""

        data = validate_type(data, 'data', np.ndarray)
        channels = validate_type(channels, 'channels', np.ndarray)

        param_list = [{"data":data[i], "channel_names":channels[i], "filtered":filtered} for i in range(len(channels))]
        ret = parallel_process(param_list, self.detect, self.n_jobs, self.use_kwargs, self.front_num)
        channel_name, HFOs = [], []
        for j in ret:
            if not type(j) is tuple:
                print(j)
            if j[0] is None:
                continue
            HFOs.append(j[0])
            channel_name.append(j[1])
        channel_name = np.array(channel_name)
        HFOs = np.array(HFOs, dtype=object)
        index = reorder(channel_name, channels)
        return channel_name[index], HFOs[index]


    def detect(self, data, channel_names, filtered=False):
        """Detect HFOs from a single channel

        Parameters
        ----------
        data : numpy array like of float, shape (n_samples,)
        channel_names : str
        filtered : bool, default False
            Whether the data is already filtered

        Returns
        -------
        HFOs : numpy array of int, shape (n_HFOs, 2)
        channel_names : str"""

        data = validate_type(data, 'data', np.ndarray)
        if data.shape[0] == 0:
            return None, None
        if data.ndim != 1:
            raise ValueError('data must be a 1D array')
        channel_names = validate_type(channel_names, 'channel_names', str)

        if filtered == False:        
            data = preprocess(data, self.sample_freq, self.filter_freq)
        rms = compute_rms(data, self.sample_freq, self.rms_window, detector='RMS')
        epoch_lims = self._compute_epoch_lims(len(data), self.sample_freq, self.epoch_len)
        HFOs = self._get_HFOs(data, rms, epoch_lims, self.sample_freq, self.min_window, self.min_gap, self.min_osc, self.rms_thres, self.peak_thres)
        return HFOs, channel_names

    
    def _compute_epoch_lims(self, data_len, sample_freq, epoch_len):
        """
        Compute the limits of each epoch
        
        Parameters
        ----------
        data_len : int | float
        sample_freq : float | int
        epoch_len : int | float

        Returns
        -------
        epoch_lims : numpy array of int, shape (n_epochs, 2)
        """

        # compute epoch limits
        epoch_len = round(epoch_len * sample_freq)
        temp = np.arange(0, data_len, epoch_len)

        if temp[-1] < data_len:
            temp = np.append(temp, [data_len])
        epoch_lims = np.vstack([[temp[:-1].T, temp[1:]]]).T
        return epoch_lims


    def _get_HFOs(self, filtered, rms, epoch_lims, sample_freq, min_window, min_gap, min_osc, rms_thres, peak_thres):
        """
        Get HFOs from a single channel
        
        Parameters
        ----------
        filtered : numpy array of float, shape (n_samples,)
        rms : numpy array of float, shape (n_samples,)
        epoch_lims : numpy array of int, shape (n_epochs, 2)
        sample_freq : float | int
        min_window : int | float
        min_gap : int | float
        min_osc : int
        rms_thres : int | float
        peak_thres : int | float

        Returns
        -------
        HFOs : numpy array of int, shape (n_HFOs, 2)
        """
        
        # check input parameters
        if filtered.ndim != 1:
            raise ValueError('filtered must be a 1D array')
        if rms.ndim != 1:
            raise ValueError('rms must be a 1D array')
        if epoch_lims.ndim != 2:    
            raise ValueError('epoch_lims must be a 2D array')   
        if epoch_lims.shape[1] != 2:
            raise ValueError('epoch_lims must have 2 columns')
        if rms.shape[0] != filtered.shape[0]:
            raise ValueError('rms and filtered must have the same length')
        
        min_window = round(min_window * sample_freq)
        min_gap = round(min_gap * sample_freq)

        # get HFOs
        HFOs = []
        for i, j in epoch_lims:
            # calculate threshold
            window = np.zeros(len(rms))
            window[i:j] = 1
            rms_epoch = rms * window
            rms_interval = rms[i:j]
            epoch_filt = filtered[i:j]
            thres_rms = (rms_epoch > (np.mean(rms_interval) + rms_thres * np.std(rms_interval))).astype('int')      

            if len(np.argwhere(thres_rms)) == 0:
                # print("none satisfies THRES_RMS requirement")
                pass

            wind_thres = np.pad(thres_rms, 1)
            wind_jumps = np.diff(wind_thres)
            wind_jump_up = np.argwhere(wind_jumps == 1)
            wind_jump_down = np.argwhere(wind_jumps == -1)
            wind_dist = wind_jump_down - wind_jump_up
            wind_ini = wind_jump_up[wind_dist > min_window] + 1  
            wind_end = wind_jump_down[wind_dist > min_window]

            if len(wind_ini) == 0:
                # print("none satisfies MIN_WINDOW requirement")
                pass

            while True:
                next_ini = wind_ini[1:]
                last_end = wind_end[:-1]
                wind_idx = (next_ini - last_end) < min_gap

                if np.sum(wind_idx) == 0:
                    # print("break while")
                    break

                new_end = wind_end[1:]
                last_end[wind_idx] = new_end[wind_idx]
                wind_end[:-1] = last_end

                idx = np.diff(np.pad(wind_end, (1, 0))) != 0
                wind_ini = wind_ini[idx]
                wind_end = wind_end[idx]    

            wind_intervals = np.array([wind_ini, wind_end]).T

            # select intervals 
            count = 1
            wind_select = []
            thres_peak = np.mean(np.abs(epoch_filt) + peak_thres * np.std(np.abs(epoch_filt)))

            for ii, jj in wind_intervals:
                temp = np.abs(filtered[ii-1:jj])
                if len(temp) < 3:
                    continue

                peak_ind, _ = find_peaks(temp, height=thres_peak)
                if len(peak_ind) < min_osc:
                    continue

                wind_select.append([ii, jj])
                count += 1

            if len(wind_select):
                HFOs += wind_select
        return HFOs                             




import numpy as np
from scipy.signal import *
from .utils import *

class MNIDetector():

    def __init__(self, sample_freq, filter_freq=[80, 500], 
                epoch_time=10, epo_CHF=60, per_CHF=95/100, 
                min_win=10*1e-3, min_gap=10*1e-3, thrd_perc=99.9999/100, 
                base_seg=125*1e-3, base_shift=0.5, base_thrd=0.67, base_min=5,
                n_jobs=32, use_kwargs=True, front_num=1, seed=None):
        """Initialize the MNI detector
        sample_freq : float | int
            Sampling frequency of the data in Hz
        filter_freq : list of float or int, length 2, default [80, 500] Hz
            Filter freqs in Hz
        epoch_time: float | int, default 10 s
            Cycle Time in seconds(s)
        epo_CHF: float | int, default 60 s
            Continuous High Frequency epoch in seconds(s)
        per_CHF: float | int, default 95 %
            Continous High Frequency percentage (%)
        min_win: float | int, default 10 * 1e-3 s
            Minimum HFO time in seconds(s)
        min_gap: float | int, default 10 * 1e-3s
            Minimum gap between HFOs in seconds(s)
        thrd_perc: float | int, default 99.9999 %
            Threshold percentile (%)
        base_seg: float | int, default 125 * 1e-3 s
            Baseline window size in seconds(s)
        base_shift: float | int, default 0.5
            Baseline window shift
        base_thrd: float | int, default 0.67
            Baseline threshold 
        base_min: float | int, default 5
            Baseline minimum time
        n_jobs : int, default 32
            Number of jobs to run in parallel
        use_kwargs : bool, default True
            Whether to pass additional arguments to the map function
        front_num : int, default 1
            Number of jobs to run in parallel at the front end
        seed : int, default None
            Random seed for reproducibility
        """
        sample_freq = validate_param(sample_freq, 'sample_freq', (int, float))
        filter_freq = validate_filter_freq(filter_freq)
        if filter_freq[1] > sample_freq/2:
            raise ValueError('filter_freq[1] must be less than sample_freq/2')
        epoch_time = validate_param(epoch_time, 'epoch_time', (int, float))
        epo_CHF = validate_param(epo_CHF, 'epo_CHF', (int, float))
        per_CHF = validate_param(per_CHF, 'per_CHF', (int, float))
        min_win = validate_param(min_win, 'min_win', (int, float))
        min_gap = validate_param(min_gap, 'min_gap', (int, float))
        thrd_perc = validate_param(thrd_perc, 'thrd_perc', (int, float))       
        base_seg = validate_param(base_seg, 'base_seg', (int, float))
        base_shift = validate_param(base_shift, 'base_shift', (int, float))
        base_thrd = validate_param(base_thrd, 'base_thr', (int, float))
        base_min = validate_param(base_min, 'base_min', (int, float))
        n_jobs = validate_param(n_jobs, 'n_jobs', int)
        use_kwargs = validate_param(use_kwargs, 'use_kwargs', bool)
        front_num = validate_param(front_num, 'front_num', int)
        #seed = validate_param(seed, 'seed', (int, type(None)))


        self.sample_freq = sample_freq
        self.filter_freq = filter_freq
        self.epoch_time = epoch_time
        self.epo_CHF = epo_CHF
        self.per_CHF = per_CHF
        self.min_win = min_win
        self.min_gap = min_gap
        self.thrd_perc = thrd_perc
        self.base_seg = base_seg
        self.base_shift = base_shift
        self.base_thrd = base_thrd
        self.base_min = base_min
        self.n_jobs = n_jobs
        self.use_kwargs = use_kwargs
        self.front_num = front_num   
        self.seed = seed 

    def detect_edf(self, edf_path):
        """Detect HFOs from an EDF file
        
        Parameters
        ----------
        edf_path : str
            Path to the EDF file
            
        Returns
        -------
        channel_names : numpy array of str, shape (n_channels,)
        HFOs : numpy array of int, shape (n_channels, n_HFOs, 2)"""
        edf_path = validate_type(edf_path, 'edf_path', str)
        raw, channels = read_raw(edf_path)
        return self.detect_multi_channels(data=raw, channels=channels) 


    def detect_multi_channels(self, data, channels, filtered=False):
        """Detect HFOs from numpy arrays of multi channels parrallelly
        
        Parameters
        ----------
        data : numpy array like of float, shape (n_channels, n_samples)
        channels : numpy array like of str, shape (n_channels,)
        filtered : bool, default False, 
            Whether the data is already filtered

        Returns
        -------
        channel_names : numpy array of str, shape (n_channels,)
        HFOs : numpy array of int, shape (n_channels, n_HFOs, 2)"""
        
        data = validate_type(data, 'data', np.ndarray)
        channels = validate_type(channels, 'channels', np.ndarray)

        param_list = [{"data":data[i], "channel_names":channels[i], "filtered":filtered} for i in range(len(channels))]
        ret = parallel_process(param_list, self.detect, self.n_jobs, self.use_kwargs, self.front_num)
        channel_name, HFOs = [], []
        for j in ret:
            if not type(j) is tuple:
                print(j)
            if j[0] is None:
                continue
            HFOs.append(j[0])
            channel_name.append(j[1])
        channel_name = np.array(channel_name)
        HFOs = np.array(HFOs, dtype=object)
        index = reorder(channel_name, channels)
        return channel_name[index], HFOs[index]


    def detect(self, data, channel_names, filtered=False):
        """Detect HFOs from a single channel

        Parameters
        ----------
        data : numpy array like of float, shape (n_samples,)
        channel_names : str

        Returns
        -------
        HFOs : numpy array of int, shape (n_HFOs, 2)
        channel_names : str"""

        data = validate_type(data, 'data', np.ndarray)
        if data.shape[0] == 0:
            return None, None
        if data.ndim != 1:
            raise ValueError('data must be a 1D array')
        channel_names = validate_type(channel_names, 'channel_names', str)
        if not filtered:
            data = preprocess(data, self.sample_freq, self.filter_freq)
        rms = compute_rms(data, self.sample_freq, detector='MNI')
        baseline_window = self._detect_baseline(data, self.sample_freq, self.filter_freq, self.base_seg, self.base_shift, self.base_thrd, self.seed)
        thrd = self._compute_thrd(len(data), rms, baseline_window, self.base_min, self.epoch_time, self.sample_freq, self.thrd_perc, self.min_win, self.epo_CHF, self.per_CHF)
        HFOs = self._get_HFOs(rms, thrd, self.min_win, self.sample_freq, self.min_gap)
        return HFOs, channel_names


    def _detect_baseline(self, filtered, sample_freq, filter_freq, base_seg, base_shift, base_thrd, seed):
        """Detect baseline segments
        
        Parameters
        ----------
        filtered : numpy array of float, shape (n_samples,)
        sample_freq : int or float
        filter_freq : list
        base_seg : float or int
        base_shift : float or int  
        base_thr : float or int
        
        Returns
        -------
        baseline_window : numpy array"""

        # check inputs
        if filtered.ndim != 1:
            raise ValueError('filtered must be a 1D array')

        # init outputs
        data_len = len(filtered)
        baseline_window = np.zeros((data_len, 1))

        # get wavelet entropy
        epoch_samples = round(sample_freq * base_seg)
        low, high = filter_freq
        wavelet_entropy = compute_wavelet_entropy(sample_freq, epoch_samples, low, high, seed)
        
        # get baseline
        in_idx = list(range(1, data_len + 1, round(epoch_samples * base_shift)))
        en_idx = [idx + epoch_samples - 1 for idx in in_idx]
        idx = [1 if i > data_len else 0 for i in en_idx]
        in_idx = in_idx[: idx.index(1)]
        en_idx = en_idx[: idx.index(1)]
        del idx

        for i in range(0, len(in_idx)):
            sig_filt = filtered[in_idx[i] - 1:en_idx[i]]
            auto_corr = correlate(sig_filt, sig_filt) / sum(sig_filt ** 2)
            w_coef = gabor_transform_wait(auto_corr, sample_freq, low, high)
            prob_energy = np.mean(w_coef ** 2, 1) / sum(np.mean(w_coef ** 2, 1))
            we_section = -sum(prob_energy * np.log(prob_energy))
            if we_section < base_thrd * wavelet_entropy:
                baseline_window[in_idx[i] - 1:en_idx[i]] = 1
        del prob_energy, w_coef
            
        return baseline_window
    

    def _compute_thrd(self, data_len, rms, baseline_window, base_min, epoch_time, sample_freq, thrd_perc, min_win, epo_CHF, per_CHF):
        """Compute threshold
        
        Parameters
        ----------
        data_len : int
            length of data
        rms : numpy array, shape (n_samples,)
            rms of data
        baseline_window : numpy array of 0 and 1, shape (n_samples, 1)
            baseline window
        base_min : float or int
            Baseline minimum time
        epoch_time : float or int
            Cycle Time
        sample_freq : int or float
            Sampling frequency
        thrd_perc : float or int
            Threshold percentile
        min_win : float or int
            Minimum HFO Time   
        epo_CHF : float or int
            Continous High Frequency Epoch
        per_CHF : float or int
            Continous High Frequency Percentil Threshold  
        
        Returns
        -------
        thrd : numpy array
        """
        # check data
        if rms.ndim != 1:
            raise ValueError('rms must be a 1D array')
        if baseline_window.ndim != 2 and baseline_window.shape[1] != 1:
            raise ValueError('baseline_window must be a 2D array with shape (n_samples, 1)')
        if not np.equal(baseline_window, 0).any() or not np.equal(baseline_window, 1).any():
            raise ValueError('baseline_window contain only 0 and 1')
        if len(rms) != data_len and len(baseline_window) != data_len:
            raise ValueError('rms and baseline_window must have the same length as data_len')

        # compute threshold
        wind_samples = (base_min / 60) * data_len 

        if np.sum(baseline_window) >= wind_samples:
            thrd = pos_baseline(baseline_window, rms, epoch_time, sample_freq, thrd_perc)
        else:
            thrd = neg_baseline(rms, min_win, sample_freq, epo_CHF, per_CHF)

        return thrd
    

    def _get_HFOs(self, rms, thrd, min_win, sample_freq, min_gap):
        """Get HFOs
        
        Parameters
        ----------
        rms : numpy array of float, shape (n_samples,)
            rms
        thrd : numpy array of float, shape (n_samples,)
            threshold
        min_win : int|float
            Minimum HFO Time     
        sample_freq : int|float
            sampling frequency
        min_gap : int|float
            Minimum HFO Gap
        
        Returns
        -------
        HFOs : numpy array of int, shape (n_HFOs, 2)
        """
        # check inputs
        if rms.ndim != 1:
            raise ValueError('rms must be a 1D array')
        if thrd.ndim != 1:
            raise ValueError('thrd must be a 1D array')
        if len(rms) != len(thrd):
            raise ValueError('rms and thrd must have the same length')
 
        # init outputs
        HFOs = []
        
        # get HFOs
        min_win = min_win * sample_freq
        energy_thrd = np.zeros(len(rms)+2)
        energy_thrd[1:-1] = rms >= thrd
        wind_jumps = np.diff(energy_thrd)
        wind_jump_up = np.where(wind_jumps == 1)[0]
        wind_jump_down = np.where(wind_jumps == -1)[0]
        wind_dist = wind_jump_down - wind_jump_up

        wind_select = np.where(wind_dist > min_win)[0]

        if wind_select.size == 0:
            HFO_events = []
            print('Event Selection - No detected')
            return 

        wind_ini = wind_jump_up[wind_select]
        wind_end = wind_jump_down[wind_select]
        min_gap = min_gap * sample_freq

        del wind_jumps, wind_jump_up, wind_jump_down, wind_select

        while True:
            if wind_ini.size == 0:
                HFO_events = []
                print('Event Selection - No detected')
                return

            if len(wind_ini) < 2:
                break

            next_ini = wind_ini[1:]
            last_end = wind_end[:-1]
            wind_idx = (next_ini - last_end) < min_gap

            if np.sum(wind_idx) == 0:
                break

            new_end = wind_end[1:]
            last_end[wind_idx] = new_end[wind_idx]
            wind_end[:-1] = last_end

            v_idx = np.zeros(len(wind_end)+1)
            v_idx[1:] = wind_end
            v_idx = np.diff(v_idx) != 0
            wind_ini = wind_ini[v_idx]
            wind_end = wind_end[v_idx]
        
        wind_ini = wind_ini.reshape(-1, 1)
        wind_end = wind_end.reshape(-1, 1)
        HFO_events = np.hstack((wind_ini + 1, wind_end + 1)).tolist()

            
        return HFO_events



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='HFODetector',
    version='0.0.16',
    description='Package for detecting HFOs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/roychowdhuryresearch/HFO_Detector',
    author='Xin Chen, Hoyoung Chung',
    author_email='xinchen98@g.ucla.edu, taylorchung@ucla.edu',
    license='MIT License',
    python_requires='>=3.6',
    packages=find_packages(include=['HFODetector']),
    install_requires=['mne',
                      'numpy',
                      'scipy',
                      'matplotlib',
                      'pandas',
                      'scikit-image'
                      ],

    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
)
